import json
import os
import re
import time
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List

import requests
from bs4 import BeautifulSoup
from cloudscraper import create_scraper

from udemy_enroller.logging import get_logger
from udemy_enroller.settings import Settings
from udemy_enroller.utils import get_app_dir

logger = get_logger()


def format_requests(func):
    """
    Convenience method for handling requests response
    """

    def formatting(*args, **kwargs):
        result = func(*args, **kwargs)
        result.raise_for_status()
        return result.json()

    return formatting


@dataclass(unsafe_hash=True)
class RunStatistics:
    prices: List[float] = field(default_factory=list)

    expired: int = 0
    enrolled: int = 0
    already_enrolled: int = 0
    unwanted_language: int = 0
    unwanted_category: int = 0

    course_ids_start: int = 0
    course_ids_end: int = 0

    start_time = None
    end_time = None

    currency_symbol = "$"

    def savings(self):
        return sum(self.prices) or 0

    def table(self):
        logger.info("================== Run Statistics ==================")
        logger.info(f"Enrolled:                   {self.enrolled}")
        logger.info(f"Unwanted Category:          {self.unwanted_category}")
        logger.info(f"Unwanted Language:          {self.unwanted_language}")
        logger.info(f"Already Claimed:            {self.already_enrolled}")
        logger.info(f"Expired:                    {self.expired}")
        logger.info(f"Total Enrolments:           {self.course_ids_end}")
        logger.info(
            f"Savings:                    {self.currency_symbol}{self.savings():.2f}"
        )
        logger.info("================== Run Statistics ==================")


class UdemyStatus(Enum):
    """
    Possible statuses of udemy course
    """

    ALREADY_ENROLLED = "ALREADY_ENROLLED"
    ENROLLED = "ENROLLED"
    EXPIRED = "EXPIRED"
    UNWANTED_LANGUAGE = "UNWANTED_LANGUAGE"
    UNWANTED_CATEGORY = "UNWANTED_CATEGORY"


class UdemyActions:
    LOGIN_URL = "https://www.udemy.com/join/login-popup/?locale=en_US"
    MY_COURSES = (
        "https://www.udemy.com/api-2.0/users/me/subscribed-courses/?ordering=-last_accessed&fields["
        "course]=@min,enrollment_time,published_title,&fields[user]=@min"
    )
    CHECKOUT_URL = "https://www.udemy.com/payment/checkout-submit/"
    CHECK_PRICE = (
        "https://www.udemy.com/api-2.0/course-landing-components/{}/me/?couponCode={"
        "}&components=price_text,deal_badge,discount_expiration"
    )
    COURSE_DETAILS = (
        "https://www.udemy.com/api-2.0/courses/{}/?fields[course]=title,context_info,primary_category,"
        "primary_subcategory,avg_rating_recent,visible_instructors,locale,estimated_content_length,"
        "num_subscribers"
    )
    USER_DETAILS = "https://www.udemy.com/api-2.0/contexts/me/?me=True&Config=True"

    HEADERS = {
        "origin": "https://www.udemy.com",
        "accept": "application/json, text/plain, */*",
        "accept-encoding": "gzip, deflate, br",
        "content-type": "application/json;charset=UTF-8",
        "x-requested-with": "XMLHttpRequest",
        "x-checkout-version": "2",
        "referer": "https://www.udemy.com/",
    }

    def __init__(self, settings: Settings, cookie_file_name: str = ".cookie"):
        self.settings = settings
        self.user_has_preferences = self.settings.categories or self.settings.languages
        self.session = requests.Session()
        self.udemy_scraper = create_scraper(ecdhCurve="secp384r1")
        self._cookie_file = os.path.join(get_app_dir(), cookie_file_name)
        self._enrolled_course_info = []
        self._all_course_ids = []
        self._currency_symbol = None
        self._currency = None

        self.stats = RunStatistics()

    def login(self, retry=False) -> None:
        """
        Login to Udemy using REST api
        Saves login cookies for future use

        :return: None
        """
        cookie_details = self._load_cookies()
        if cookie_details is None:
            response = self.udemy_scraper.get(self.LOGIN_URL)
            soup = BeautifulSoup(response.content, "html.parser")
            csrf_element = soup.find("input", {"name": "csrfmiddlewaretoken"}) or {}
            csrf_token = csrf_element.get("value")
            if csrf_token is None:
                raise Exception("Unable to get csrf_token")

            # Prompt for email/password if we don't have them saved in settings
            if self.settings.email is None:
                self.settings.prompt_email()
            if self.settings.password is None:
                self.settings.prompt_password()

            _form_data = {
                "email": self.settings.email,
                "password": self.settings.password,
                "csrfmiddlewaretoken": csrf_token,
            }
            self.udemy_scraper.headers.update({"Referer": self.LOGIN_URL})
            auth_response = self.udemy_scraper.post(
                self.LOGIN_URL, data=_form_data, allow_redirects=False
            )
            if auth_response.status_code != 302:
                logger.debug(
                    f"Error while trying to login: {auth_response.status_code}"
                )
                logger.debug(f"Failed login response: {auth_response.text}")
                raise Exception(f"Could not login. Code: {auth_response.status_code}")
            else:
                cookie_details = {
                    "csrf_token": csrf_token,
                    "access_token": auth_response.cookies["access_token"],
                    "client_id": auth_response.cookies["client_id"],
                }
                self._cache_cookies(cookie_details)

        bearer_token = f"Bearer {cookie_details['access_token']}"
        self.session.headers = self.HEADERS
        self.session.headers.update(
            {
                "authorization": bearer_token,
                "x-udemy-authorization": bearer_token,
                "x-csrftoken": cookie_details["csrf_token"],
            }
        )
        self.session.cookies.update({"access_token": cookie_details["access_token"]})
        self.session.cookies.update({"client_id": cookie_details["client_id"]})

        try:
            self._enrolled_course_info = self.load_my_courses()

            user_details = self.load_user_details()
            # Extract the users currency info needed for checkout
            self._currency = user_details["Config"]["price_country"]["currency"]
            self._currency_symbol = user_details["Config"]["price_country"][
                "currency_symbol"
            ]
            self._all_course_ids = [
                course["id"] for course in self._enrolled_course_info
            ]
            self.stats.course_ids_start = len(self._all_course_ids)
            self.stats.currency_symbol = self._currency_symbol
        except Exception as e:
            # Log some info on the HTTPError we are getting
            if isinstance(e, requests.HTTPError):
                logger.error("HTTP error while trying to fetch Udemy information")
                logger.error(e)
                retry = True
            if not retry:
                logger.info("Retrying login")
                self._delete_cookies()
                return self.login(retry=True)
            else:
                logger.error(
                    "Unable to fetch existing courses. Login was not successful"
                )
                raise e

    def load_my_courses(self) -> List:
        """
        Loads users currently enrolled courses from Udemy

        :return: List of logged in users courses
        """
        logger.info("Loading existing course details")
        all_courses = list()
        page_size = 100

        my_courses = self.my_courses(1, page_size)
        all_courses.extend(my_courses["results"])
        total_pages = my_courses["count"] // page_size
        for page in range(2, total_pages + 2):
            my_courses = self.my_courses(page, page_size)
            if "results" in my_courses:
                all_courses.extend(my_courses["results"])
            time.sleep(1)
        logger.info(f"Currently enrolled in {len(all_courses)} courses")
        return all_courses

    @format_requests
    def load_user_details(self):
        """
        Load the current users details

        :return: Dict containing the users details
        """
        return self.session.get(self.USER_DETAILS)

    def is_enrolled(self, course_id: int) -> bool:
        """
        Check if the user is currently enrolled in the course based on course_id passed in

        :param int course_id: Check if the course_id is in the users current courses
        :return:
        """
        return course_id in self._all_course_ids

    def _add_enrolled_course(self, course_id):
        """
        Add enrolled course to the list of enrolled course ids

        :param int course_id: The course_id to add to the list
        :return:
        """
        self._all_course_ids.append(course_id)
        self.stats.course_ids_end = len(self._all_course_ids)

    def is_coupon_valid(
        self, course_id: int, coupon_code: str, course_identifier: str
    ) -> bool:
        """
        Check if the coupon is valid for a course

        :param int course_id: Id of the course to check the coupon against
        :param str coupon_code: Coupon to apply to the course
        :param str course_identifier: Name of the course used for logging
        :return:
        """
        coupon_valid = True
        coupon_details = self.coupon_details(course_id, coupon_code)
        current_price = coupon_details["price_text"]["data"]["pricing_result"]["price"][
            "amount"
        ]
        if bool(current_price):
            logger.debug(
                f"Skipping course '{course_identifier}' as it now costs {self._currency_symbol}{current_price}"
            )
            coupon_valid = False
        if not bool(
            coupon_details["price_text"]["data"]["pricing_result"]["list_price"][
                "amount"
            ]
        ):
            logger.debug(f"Skipping course '{course_identifier}' as it is always FREE")
            coupon_valid = False

        if coupon_valid:
            usual_price = coupon_details["price_text"]["data"]["pricing_result"][
                "saving_price"
            ]["amount"]
            self.stats.prices.append(usual_price)
        return coupon_valid

    def is_preferred_language(
        self, course_details: Dict, course_identifier: str
    ) -> bool:
        """
        Check if the course is in one of the languages preferred by the user

        :param dict course_details: Dictionary containing course details from Udemy
        :param str course_identifier: Name of the course used for logging
        :return: boolean
        """
        is_preferred_language = True
        course_language = course_details["locale"]["simple_english_title"]
        if course_language not in self.settings.languages:
            logger.debug(
                f"Course '{course_identifier}' language not wanted: {course_language}"
            )
            is_preferred_language = False

        return is_preferred_language

    def is_preferred_category(
        self, course_details: Dict, course_identifier: str
    ) -> bool:
        """
        Check if the course is in one of the categories preferred by the user

        :param dict course_details: Dictionary containing course details from Udemy
        :param str course_identifier: Name of the course used for logging
        :return: boolean
        """
        is_preferred_category = True

        if (
            course_details["primary_category"]["title"] not in self.settings.categories
            and course_details["primary_subcategory"]["title"]
            not in self.settings.categories
        ):
            logger.debug(
                f"Skipping course '{course_identifier}' as it does not have a wanted category"
            )
            is_preferred_category = False
        return is_preferred_category

    @format_requests
    def my_courses(self, page: int, page_size: int) -> Dict:
        """
        Load the current logged in users courses

        :param int page: page number to load
        :param int page_size: number of courses to load per page
        :return: dict containing the current users courses
        """
        return self.session.get(self.MY_COURSES + f"&page={page}&page_size={page_size}")

    @format_requests
    def coupon_details(self, course_id: int, coupon_code: str) -> Dict:
        """
        Check that the coupon is valid for the current course

        :param int course_id: Id of the course to check the coupon against
        :param str coupon_code: The coupon_code to check against the course
        :return: dictionary containing the course pricing details
        """
        return requests.get(self.CHECK_PRICE.format(course_id, coupon_code))

    @format_requests
    def course_details(self, course_id: int) -> Dict:
        """
        Retrieves details relating to the course passed in

        :param int course_id: Id of the course to get the details of
        :return: dictionary containing the course details
        """
        return requests.get(self.COURSE_DETAILS.format(course_id))

    def enroll(self, course_link: str) -> str:
        """
        Enroll the current user in the course provided

        :param str course_link: Link to the course with valid coupon attached
        :return: str representing the status of the enrolment
        """
        str_check = "?couponCode="
        if str_check in course_link:
            url, coupon_code = course_link.split(str_check)
            course_id = self._get_course_id(url)
            course_details = self.course_details(course_id)
            course_identifier = course_details.get("title", url)

            if self.is_enrolled(course_id):
                logger.info(f"Already enrolled in: '{course_identifier}'")
                self.stats.already_enrolled += 1
                return UdemyStatus.ALREADY_ENROLLED.value

            if self.user_has_preferences:
                if self.settings.languages:
                    if not self.is_preferred_language(
                        course_details, course_identifier
                    ):
                        self.stats.unwanted_language += 1
                        return UdemyStatus.UNWANTED_LANGUAGE.value
                if self.settings.categories:
                    if not self.is_preferred_category(
                        course_details, course_identifier
                    ):
                        self.stats.unwanted_category += 1
                        return UdemyStatus.UNWANTED_CATEGORY.value

            if not self.is_coupon_valid(course_id, coupon_code, course_identifier):
                self.stats.expired += 1
                return UdemyStatus.EXPIRED.value

            return self._checkout(course_id, coupon_code, course_identifier)
        else:
            logger.debug(f"Malformed url passed in: {course_link}")
            return UdemyStatus.EXPIRED.value

    def _get_course_id(self, url: str) -> int:
        """
        Get the course id from the url provided

        :param str url: Udemy url to fetch the course from
        :return: int representing the course id
        """
        response = self.session.get(url)
        response.raise_for_status()
        soup = BeautifulSoup(response.content, "html.parser")

        return int(soup.find("body")["data-clp-course-id"])

    def _checkout(
        self,
        course_id: int,
        coupon_code: str,
        course_identifier: str,
        retry: bool = False,
    ) -> str:
        """
        Checkout process for the course and coupon provided

        :param int course_id: The course id of the course to enroll in
        :param str coupon_code: The coupon code to apply on checkout
        :param str course_identifier: Name of the course being checked out
        :param str retry: If this is a retried checkout raise exception if not successful
        :return:
        """
        payload = self._build_checkout_payload(course_id, coupon_code)
        checkout_result = self.session.post(self.CHECKOUT_URL, json=payload)
        if not checkout_result.ok:
            if not retry:
                seconds = int(re.search("\\d+", checkout_result.text).group()) + 1
                logger.info(
                    f"Script has been rate limited. Sleeping for {seconds} seconds"
                )
                time.sleep(seconds)
                self._checkout(course_id, coupon_code, course_identifier, retry=True)
            else:
                raise Exception(
                    f"Checkout failed: Code: {checkout_result.status_code} Text: {checkout_result.text}"
                )
        else:
            result = checkout_result.json()
            if result["status"] == "succeeded":
                logger.info(f"Successfully enrolled: '{course_identifier}'")
                self._add_enrolled_course(course_id)
                self.stats.enrolled += 1
                return UdemyStatus.ENROLLED.value
            elif result["status"] == "failed":
                logger.warning(f"Checkout failed: '{course_identifier}'")
                logger.debug(f"Checkout payload: {payload}")
                # TODO: Shouldn't happen. Need to monitor if it does
                return UdemyStatus.EXPIRED.value

    def _build_checkout_payload(self, course_id: int, coupon_code: str) -> Dict:
        """
        Build the payload for checkout

        :param int course_id: The course id to checkout
        :param str coupon_code: The coupon code to use at checkout
        :return: dict representing the checkout payload
        """
        return {
            "checkout_event": "Submit",
            "checkout_environment": "Marketplace",
            "shopping_info": {
                "items": [
                    {
                        "discountInfo": {"code": coupon_code},
                        "buyable": {"type": "course", "id": course_id, "context": {}},
                        "price": {"amount": 0, "currency": self._currency},
                    }
                ],
                "is_cart": True,
            },
            "payment_info": {"payment_vendor": "Free", "payment_method": "free-method"},
        }

    def _cache_cookies(self, cookies: Dict) -> None:
        """
        Caches cookies for future logins

        :param cookies:
        :return:
        """
        logger.info("Caching cookie for future use")
        with open(self._cookie_file, "a+") as f:
            f.write(json.dumps(cookies))

    def _load_cookies(self) -> Dict:
        """
        Loads existing cookie file

        :return:
        """
        cookies = None

        if os.path.isfile(self._cookie_file):
            logger.info("Loading cookie from file")
            with open(self._cookie_file) as f:
                cookies = json.loads(f.read())
        else:
            logger.info("No cookie available")
        return cookies

    def _delete_cookies(self) -> None:
        """
        Remove existing cookie file

        :return:
        """
        logger.info("Deleting cookie")
        os.remove(self._cookie_file)



import asyncio
import logging
from typing import List

from bs4 import BeautifulSoup

from udemy_enroller.http import get
from udemy_enroller.scrapers.base_scraper import BaseScraper

logger = logging.getLogger("udemy_enroller")


class TutorialBarScraper(BaseScraper):
    """
    Contains any logic related to scraping of data from tutorialbar.com
    """

    DOMAIN = "https://www.tutorialbar.com"
    AD_DOMAINS = ("https://amzn", "https://bit.ly")

    def __init__(self, enabled, max_pages=None):
        super().__init__()
        self.scraper_name = "tutorialbar"
        if not enabled:
            self.set_state_disabled()
        self.last_page = None
        self.max_pages = max_pages

    @BaseScraper.time_run
    async def run(self) -> List:
        """
        Runs the steps to scrape links from tutorialbar.com

        :return: list of udemy coupon links
        """
        links = await self.get_links()
        self.max_pages_reached()
        return links

    async def get_links(self):
        """
        Scrape udemy links from tutorialbar.com

        :return: List of udemy course urls
        """
        self.current_page += 1
        course_links = await self.get_course_links(
            f"{self.DOMAIN}/all-courses/page/{self.current_page}/"
        )

        logger.info(
            f"Page: {self.current_page} of {self.last_page} scraped from tutorialbar.com"
        )
        udemy_links = await self.gather_udemy_course_links(course_links)
        links = self._filter_ad_domains(udemy_links)

        for counter, course in enumerate(links):
            logger.debug(f"Received Link {counter + 1} : {course}")

        return links

    def _filter_ad_domains(self, udemy_links) -> List:
        """
        Filter out any known ad domains from the links scraped

        :param list udemy_links: List of urls to filter ad domains from
        :return: A list of filtered urls
        """
        ad_links = set()
        for link in udemy_links:
            for ad_domain in self.AD_DOMAINS:
                if link.startswith(ad_domain):
                    ad_links.add(link)
        if ad_links:
            logger.info(f"Removing ad links from courses: {ad_links}")
        return list(set(udemy_links) - ad_links)

    async def get_course_links(self, url: str) -> List:
        """
        Gets the url of pages which contain the udemy link we want to get

        :param str url: The url to scrape data from
        :return: list of pages on tutorialbar.com that contain Udemy coupons
        """
        text = await get(url)
        if text is not None:
            soup = BeautifulSoup(text.decode("utf-8"), "html.parser")

            links = soup.find_all("h3")
            course_links = [link.find("a").get("href") for link in links]

            self.last_page = (
                soup.find("li", class_="next_paginate_link")
                .find_previous_sibling()
                .text
            )

            return course_links

    @staticmethod
    async def get_udemy_course_link(url: str) -> str:
        """
        Gets the udemy course link

        :param str url: The url to scrape data from
        :return: Coupon link of the udemy course
        """

        text = await get(url)
        if text is not None:
            soup = BeautifulSoup(text.decode("utf-8"), "html.parser")
            udemy_link = (
                soup.find("span", class_="rh_button_wrapper").find("a").get("href")
            )
            return udemy_link

    async def gather_udemy_course_links(self, courses: List[str]):
        """
        Async fetching of the udemy course links from tutorialbar.com

        :param list courses: A list of tutorialbar.com course links we want to fetch the udemy links for
        :return: list of udemy links
        """
        return [
            link
            for link in await asyncio.gather(*map(self.get_udemy_course_link, courses))
            if link is not None
        ]



import getpass
import os.path
from distutils.util import strtobool
from typing import Dict, List, Tuple

from ruamel.yaml import YAML, dump

from udemy_enroller.logging import get_logger
from udemy_enroller.utils import get_app_dir

logger = get_logger()


class Settings:
    """
    Contains all logic related to the scripts settings
    """

    def __init__(
        self, delete_settings=False, delete_cookie=False, settings_path="settings.yaml"
    ):
        self.email = None
        self.password = None
        self.zip_code = None
        self.languages = []
        self.categories = []

        self._settings_path = os.path.join(get_app_dir(), settings_path)
        self._cookies_path = os.path.join(get_app_dir(), ".cookie")
        self._should_store_email = False
        self._should_store_password = False
        self.is_ci_build = strtobool(os.environ.get("CI_TEST", "False"))
        if delete_settings:
            self.delete_settings()
        if delete_cookie:
            self.delete_cookie()
        self._init_settings()

    def _init_settings(self) -> None:
        """
        Initialize the settings to be used in the script

        :return:
        """
        if self.is_ci_build:
            self._load_ci_settings()
        else:
            settings = self._load_user_settings()
            if settings is None:
                self._generate_settings()
                self._save_settings()

    def _load_ci_settings(self):
        """
        Load environment variables for CI run

        :return:
        """
        logger.info("Loading CI settings")
        self.email = os.environ["UDEMY_EMAIL"]
        self.password = os.environ["UDEMY_PASSWORD"]

    def _load_user_settings(self) -> Dict:
        """
        Loads the settings from the yaml file if it exists

        :return: dictionary containing the script settings
        """
        yaml = YAML()

        settings = None
        if os.path.isfile(self._settings_path):
            logger.info("Loading existing settings")
            with open(self._settings_path) as f:
                settings = yaml.load(f)
            udemy_settings = settings["udemy"]
            self.email = udemy_settings["email"]
            self.password = udemy_settings["password"]
            self.zip_code = udemy_settings.get("zipcode")
            self.languages = udemy_settings.get("languages")
            self.categories = udemy_settings.get("categories")

        return settings

    def _generate_settings(self) -> None:
        """
        Generate the settings for the script

        :return:
        """
        self.email, self._should_store_email = self._get_email()
        self.password, self._should_store_password = self._get_password()
        self.zip_code = self._get_zip_code()
        self.languages = self._get_languages()
        self.categories = self._get_categories()

    def _get_email(self, prompt_save=True) -> Tuple[str, bool]:
        """
        Get input from user on the email to use for udemy

        :return: The users udemy email and if it should be saved
        """
        email = input("Please enter your udemy email address: ")
        if len(email) == 0:
            logger.warning("You must provide your email")
            return self._get_email()
        if prompt_save:
            save_email = input("Do you want to save your email for future use (Y/N): ")
            should_store = save_email.lower() == "y"
        else:
            should_store = False
        return email, should_store

    def _get_password(self, prompt_save=True) -> Tuple[str, bool]:
        """
        Get input from user on the password to use for udemy

        :return: The users udemy password and if it should be saved
        """
        password = getpass.getpass(prompt="Please enter your udemy password: ")
        if len(password) == 0:
            logger.warning("You must provide your password")
            return self._get_password()
        if prompt_save:
            save_password = input(
                "Do you want to save your password for future use (Y/N): "
            )
            should_store = save_password.lower() == "y"
        else:
            should_store = False
        return password, should_store

    @staticmethod
    def _get_zip_code() -> str:
        """
        Get input from user on the zip code to use for udemy

        :return: The users udemy zip code
        """
        zip_code = input("Please enter your zipcode (Not necessary in some regions): ")
        return zip_code

    @staticmethod
    def _get_languages() -> List[str]:
        """
        Get input from user on the languages they want to get courses in

        :return: list of languages the user wants to redeem udemy courses in
        """
        languages = input(
            "Please enter your language preferences (comma separated list e.g. English,German): "
        )
        return [lang.strip() for lang in languages.split(",")] if languages else []

    @staticmethod
    def _get_categories() -> List[str]:
        """Gets the categories the user wants.

        :return: list of categories the user wants."""
        categories = input(
            "Please enter in a list of comma separated values of"
            " the course categories you like, for example:\n"
            "Development, Design\n> "
        )
        return (
            [category.strip() for category in categories.split(",")]
            if categories
            else []
        )

    def _save_settings(self) -> None:
        """
        Confirm if the user wants to save settings to file

        :return:
        """
        yaml_structure = {
            "udemy": {
                "email": str(self.email) if self._should_store_email else None,
                "password": str(self.password) if self._should_store_password else None,
                "zipcode": str(self.zip_code),
                "languages": self.languages,
                "categories": self.categories,
            }
        }

        with open(self._settings_path, "w+") as f:
            dump(yaml_structure, stream=f)
        logger.info(f"Saved your settings in {self._settings_path}")

        # Log some details for the user
        if not self._should_store_email:
            logger.info(f"Your email has not been saved to settings.")
        if not self._should_store_password:
            logger.info("Your password has not been saved to settings.")
        if not self._should_store_email or not self._should_store_password:
            logger.info(
                "You will be prompted to enter your email/password again when the cookie expires"
            )

    def delete_settings(self) -> None:
        """
        Delete the settings file

        :return: None
        """
        if os.path.isfile(self._settings_path):
            delete_settings = input(
                "Please confirm that you want to delete your saved settings (Y/N): "
            )
            if delete_settings.lower() == "y":
                os.remove(self._settings_path)
                logger.info(f"Settings file deleted: {self._settings_path}")
        else:
            logger.info("No settings to delete")

    def delete_cookie(self) -> None:
        """
        Delete the cookie file

        :return: None
        """
        if os.path.isfile(self._cookies_path):
            os.remove(self._cookies_path)
            logger.info(f"Cookie file deleted: {self._cookies_path}")
        else:
            logger.info("No cookie file to delete")

    def prompt_email(self) -> None:
        """
        Prompt for Udemy email only. Does not prompt for saving

        :return: None
        """
        self.email, _ = self._get_email(prompt_save=False)

    def prompt_password(self) -> None:
        """
        Prompt for Udemy password only. Does not prompt for saving

        :return: None
        """
        self.password, _ = self._get_password(prompt_save=False)



import time
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import List

from price_parser import Price
from selenium.common.exceptions import NoSuchElementException, TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver, WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from udemy_enroller.exceptions import LoginException, RobotException
from udemy_enroller.logging import get_logger
from udemy_enroller.settings import Settings

logger = get_logger()


@dataclass(unsafe_hash=True)
class RunStatistics:
    prices: List[Decimal] = field(default_factory=list)

    expired: int = 0
    enrolled: int = 0
    already_enrolled: int = 0
    unwanted_language: int = 0
    unwanted_category: int = 0

    start_time = None

    currency_symbol = None

    def savings(self):
        return sum(self.prices) or 0

    def table(self):
        # Only show the table if we have something to show
        if self.prices:
            if self.currency_symbol is None:
                self.currency_symbol = "¤"
            run_time_seconds = int(
                (datetime.utcnow() - self.start_time).total_seconds()
            )

            logger.info("==================Run Statistics==================")
            logger.info(f"Enrolled:                   {self.enrolled}")
            logger.info(f"Unwanted Category:          {self.unwanted_category}")
            logger.info(f"Unwanted Language:          {self.unwanted_language}")
            logger.info(f"Already Claimed:            {self.already_enrolled}")
            logger.info(f"Expired:                    {self.expired}")
            logger.info(
                f"Savings:                    {self.currency_symbol}{self.savings():.2f}"
            )
            logger.info(f"Total run time (seconds):   {run_time_seconds}s")
            logger.info("==================Run Statistics==================")


class UdemyStatus(Enum):
    """
    Possible statuses of udemy course
    """

    ALREADY_ENROLLED = "ALREADY_ENROLLED"
    ENROLLED = "ENROLLED"
    EXPIRED = "EXPIRED"
    UNWANTED_LANGUAGE = "UNWANTED_LANGUAGE"
    UNWANTED_CATEGORY = "UNWANTED_CATEGORY"


class UdemyActionsUI:
    """
    Contains any logic related to interacting with udemy website
    """

    DOMAIN = "https://www.udemy.com"

    def __init__(self, driver: WebDriver, settings: Settings):
        self.driver = driver
        self.settings = settings
        self.logged_in = False
        self.stats = RunStatistics()
        self.stats.start_time = datetime.utcnow()

    def login(self, is_retry=False) -> None:
        """
        Login to your udemy account

        :param bool is_retry: Is this is a login retry and we still have captcha raise RobotException

        :return: None
        """
        if not self.logged_in:
            self.driver.get(f"{self.DOMAIN}/join/login-popup/")

            # Prompt for email/password if we don't have them saved in settings
            if self.settings.email is None:
                self.settings.prompt_email()
            if self.settings.password is None:
                self.settings.prompt_password()

            try:
                email_element = self.driver.find_element_by_name("email")
                email_element.send_keys(self.settings.email)

                password_element = self.driver.find_element_by_name("password")
                password_element.send_keys(self.settings.password)

                self.driver.find_element_by_name("submit").click()
            except NoSuchElementException as e:
                is_robot = self._check_if_robot()
                if is_robot and not is_retry:
                    input(
                        "Before login. Please solve the captcha before proceeding. Hit enter once solved "
                    )
                    self.login(is_retry=True)
                    return
                if is_robot and is_retry:
                    raise RobotException("I am a bot!")
                raise e
            else:
                user_dropdown_xpath = "//a[@data-purpose='user-dropdown']"
                try:
                    WebDriverWait(self.driver, 10).until(
                        EC.presence_of_element_located((By.XPATH, user_dropdown_xpath))
                    )
                except TimeoutException:
                    is_robot = self._check_if_robot()
                    if is_robot and not is_retry:
                        input(
                            "After login. Please solve the captcha before proceeding. Hit enter once solved "
                        )
                        if self._check_if_robot():
                            raise RobotException("I am a bot!")
                        self.logged_in = True
                        return
                    raise LoginException("Udemy user failed to login")
                self.logged_in = True

    def enroll(self, url: str) -> str:
        """
        Redeems the course url passed in

        :param str url: URL of the course to redeem
        :return: A string detailing course status
        """
        self.driver.get(url)

        course_name = self.driver.title

        if not self._check_languages(course_name):
            return UdemyStatus.UNWANTED_LANGUAGE.value

        if not self._check_categories(course_name):
            return UdemyStatus.UNWANTED_CATEGORY.value

        # TODO: Make this depend on an element.
        time.sleep(2)

        # Enroll Now 1
        buy_course_button_xpath = "//button[@data-purpose='buy-this-course-button']"
        # We need to wait for this element to be clickable before checking if already purchased
        WebDriverWait(self.driver, 10).until(
            EC.element_to_be_clickable((By.XPATH, buy_course_button_xpath))
        )

        # Check if already enrolled. If add to cart is available we have not yet enrolled
        if not self._check_enrolled(course_name):
            element_present = EC.presence_of_element_located(
                (By.XPATH, buy_course_button_xpath)
            )
            WebDriverWait(self.driver, 10).until(element_present).click()

            # Enroll Now 2
            enroll_button_xpath = "//div[starts-with(@class, 'checkout-button--checkout-button--container')]//button"
            element_present = EC.presence_of_element_located(
                (
                    By.XPATH,
                    enroll_button_xpath,
                )
            )
            WebDriverWait(self.driver, 10).until(element_present)

            # Check if zipcode exists before doing this
            if self.settings.zip_code:
                # zipcode is only required in certain regions (e.g USA)
                try:
                    element_present = EC.presence_of_element_located(
                        (
                            By.ID,
                            "billingAddressSecondaryInput",
                        )
                    )
                    WebDriverWait(self.driver, 5).until(element_present).send_keys(
                        self.settings.zip_code
                    )

                    # After you put the zip code in, the page refreshes itself and disables the enroll button for a split
                    # second.
                    enroll_button_is_clickable = EC.element_to_be_clickable(
                        (By.XPATH, enroll_button_xpath)
                    )
                    WebDriverWait(self.driver, 5).until(enroll_button_is_clickable)
                except (TimeoutException, NoSuchElementException):
                    pass

            # Make sure the price has loaded
            price_class_loading = "udi-circle-loader"
            WebDriverWait(self.driver, 10).until_not(
                EC.presence_of_element_located((By.CLASS_NAME, price_class_loading))
            )

            # Make sure the course is Free
            if not self._check_price(course_name):
                return UdemyStatus.EXPIRED.value

            # Check if state/province element exists
            billing_state_element_id = "billingAddressSecondarySelect"
            billing_state_elements = self.driver.find_elements_by_id(
                billing_state_element_id
            )
            if billing_state_elements:
                # If we are here it means a state/province element exists and needs to be filled
                # Open the dropdown menu
                billing_state_elements[0].click()

                # Pick the first element in the state/province dropdown
                first_state_xpath = (
                    "//select[@id='billingAddressSecondarySelect']//option[2]"
                )
                element_present = EC.presence_of_element_located(
                    (By.XPATH, first_state_xpath)
                )
                WebDriverWait(self.driver, 10).until(element_present).click()

            # Hit the final Enroll now button
            enroll_button_is_clickable = EC.element_to_be_clickable(
                (By.XPATH, enroll_button_xpath)
            )
            WebDriverWait(self.driver, 10).until(enroll_button_is_clickable).click()

            # Wait for success page to load
            success_element_class = (
                "//div[contains(@class, 'success-alert-banner-container')]"
            )
            WebDriverWait(self.driver, 10).until(
                EC.presence_of_element_located((By.XPATH, success_element_class))
            )

            logger.info(f"Successfully enrolled in: '{course_name}'")
            self.stats.enrolled += 1
            return UdemyStatus.ENROLLED.value
        else:
            return UdemyStatus.ALREADY_ENROLLED.value

    def _check_enrolled(self, course_name):
        add_to_cart_xpath = (
            "//div[starts-with(@class, 'buy-box')]//div[@data-purpose='add-to-cart']"
        )
        add_to_cart_elements = self.driver.find_elements_by_xpath(add_to_cart_xpath)
        if not add_to_cart_elements or (
            add_to_cart_elements and not add_to_cart_elements[0].is_displayed()
        ):
            logger.debug(f"Already enrolled in '{course_name}'")
            self.stats.already_enrolled += 1
            return True
        return False

    def _check_languages(self, course_identifier):
        is_valid_language = True
        if self.settings.languages:
            locale_xpath = "//div[@data-purpose='lead-course-locale']"
            element_text = (
                WebDriverWait(self.driver, 10)
                .until(EC.presence_of_element_located((By.XPATH, locale_xpath)))
                .text
            )

            if element_text not in self.settings.languages:
                logger.debug(f"Course language not wanted: {element_text}")
                logger.debug(
                    f"Course '{course_identifier}' language not wanted: {element_text}"
                )
                self.stats.unwanted_language += 1
                is_valid_language = False
        return is_valid_language

    def _check_categories(self, course_identifier):
        is_valid_category = True
        if self.settings.categories:
            # If the wanted categories are specified, get all the categories of the course by
            # scraping the breadcrumbs on the top

            breadcrumbs_path = "udlite-breadcrumb"
            breadcrumbs_text_path = "udlite-heading-sm"
            breadcrumbs: WebElement = self.driver.find_element_by_class_name(
                breadcrumbs_path
            )
            breadcrumbs = breadcrumbs.find_elements_by_class_name(breadcrumbs_text_path)
            breadcrumb_text = [bc.text for bc in breadcrumbs]  # Get only the text

            for category in self.settings.categories:
                if category in breadcrumb_text:
                    is_valid_category = True
                    break
            else:
                logger.debug(
                    f"Skipping course '{course_identifier}' as it does not have a wanted category"
                )
                self.stats.unwanted_category += 1
                is_valid_category = False
        return is_valid_category

    def _check_price(self, course_name):
        course_is_free = True
        price_xpath = "//div[contains(@data-purpose, 'total-amount-summary')]//span[2]"
        price_element = self.driver.find_element_by_xpath(price_xpath)

        # We are only interested in the element which is displaying the price details
        if price_element.is_displayed():
            _price = price_element.text
            # This logic should work for different locales and currencies
            checkout_price = Price.fromstring(_price)

            # Set the currency for stats
            if (
                self.stats.currency_symbol is None
                and checkout_price.currency is not None
            ):
                self.stats.currency_symbol = checkout_price.currency

            if checkout_price.amount is None or checkout_price.amount > 0:
                logger.debug(
                    f"Skipping course '{course_name}' as it now costs {_price}"
                )
                self.stats.expired += 1
                course_is_free = False

        # Get the listed price of the course for stats
        if course_is_free:
            list_price_xpath = (
                "//div[starts-with(@class, 'order-summary--original-price-text')]//span"
            )
            list_price_element = self.driver.find_element_by_xpath(list_price_xpath)
            list_price = Price.fromstring(list_price_element.text)
            if list_price.amount is not None:
                self.stats.prices.append(list_price.amount)
        return course_is_free

    def _check_if_robot(self) -> bool:
        """
        Simply checks if the captcha element is present on login if email/password elements are not

        :return: Bool
        """
        is_robot = True
        try:
            self.driver.find_element_by_id("px-captcha")
        except NoSuchElementException:
            is_robot = False
        return is_robot



from .driver_manager import ALL_VALID_BROWSER_STRINGS, DriverManager
from .logging import load_logging_config
from .scrapers.manager import ScraperManager
from .settings import Settings
from .udemy_rest import UdemyActions, UdemyStatus
from .udemy_ui import UdemyActionsUI

load_logging_config()



import argparse
import logging
from argparse import Namespace
from typing import Tuple, Union

from udemy_enroller import ALL_VALID_BROWSER_STRINGS, DriverManager, Settings
from udemy_enroller.logging import get_logger
from udemy_enroller.runner import redeem_courses, redeem_courses_ui

logger = get_logger()


def enable_debug_logging() -> None:
    """
    Enable debug logging for the scripts

    :return: None
    """
    logger.setLevel(logging.DEBUG)
    for handler in logger.handlers:
        handler.setLevel(logging.DEBUG)
    logger.info(f"Enabled debug logging")


def determine_if_scraper_enabled(
    freebiesglobal_enabled: bool,
    tutorialbar_enabled: bool,
    discudemy_enabled: bool,
    coursevania_enabled: bool,
) -> Tuple[bool, bool, bool, bool]:
    """
    Determine what scrapers should be enabled and disabled

    :return: tuple containing boolean of what scrapers should run
    """
    if (
        not freebiesglobal_enabled
        and not tutorialbar_enabled
        and not discudemy_enabled
        and not coursevania_enabled
    ):
        # Set all to True
        (
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
        ) = (True, True, True, True)

    return (
        freebiesglobal_enabled,
        tutorialbar_enabled,
        discudemy_enabled,
        coursevania_enabled,
    )


def run(
    browser: str,
    freebiesglobal_enabled: bool,
    tutorialbar_enabled: bool,
    discudemy_enabled: bool,
    coursevania_enabled: bool,
    max_pages: Union[int, None],
    delete_settings: bool,
    delete_cookie: bool,
):
    """
    Run the udemy enroller script

    :param str browser: Name of the browser we want to create a driver for
    :param bool freebiesglobal_enabled:
    :param bool tutorialbar_enabled:
    :param bool discudemy_enabled:
    :param bool coursevania_enabled:
    :param int max_pages: Max pages to scrape from sites (if pagination exists)
    :param bool delete_settings: Determines if we should delete old settings file
    :param bool delete_cookie: Determines if we should delete the cookie file
    :return:
    """
    settings = Settings(delete_settings, delete_cookie)
    if browser:
        dm = DriverManager(browser=browser, is_ci_build=settings.is_ci_build)
        redeem_courses_ui(
            dm.driver,
            settings,
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
            max_pages,
        )
    else:
        redeem_courses(
            settings,
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
            max_pages,
        )


def parse_args() -> Namespace:
    """
    Parse args from the CLI or use the args passed in

    :return: Args to be used in the script
    """
    parser = argparse.ArgumentParser(description="Udemy Enroller")

    parser.add_argument(
        "--browser",
        required=False,
        type=str,
        choices=ALL_VALID_BROWSER_STRINGS,
        help="Browser to use for Udemy Enroller",
    )
    parser.add_argument(
        "--freebiesglobal",
        action="store_true",
        default=False,
        help="Run freebiesglobal scraper",
    )
    parser.add_argument(
        "--tutorialbar",
        action="store_true",
        default=False,
        help="Run tutorialbar scraper",
    )

    parser.add_argument(
        "--discudemy",
        action="store_true",
        default=False,
        help="Run discudemy scraper",
    )

    parser.add_argument(
        "--coursevania",
        action="store_true",
        default=False,
        help="Run coursevania scraper",
    )

    parser.add_argument(
        "--max-pages",
        type=int,
        default=5,
        help=f"Max pages to scrape from sites (if pagination exists) (Default is 5)",
    )

    parser.add_argument(
        "--delete-settings",
        action="store_true",
        default=False,
        help="Delete any existing settings file",
    )

    parser.add_argument(
        "--delete-cookie",
        action="store_true",
        default=False,
        help="Delete existing cookie file",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging",
    )

    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    if args:
        if args.debug:
            enable_debug_logging()
        (
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
        ) = determine_if_scraper_enabled(
            args.freebiesglobal, args.tutorialbar, args.discudemy, args.coursevania
        )
        run(
            args.browser,
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
            args.max_pages,
            args.delete_settings,
            args.delete_cookie,
        )



import asyncio
import logging
from typing import List

from bs4 import BeautifulSoup

from udemy_enroller.http import get
from udemy_enroller.scrapers.base_scraper import BaseScraper

logger = logging.getLogger("udemy_enroller")


class DiscUdemyScraper(BaseScraper):
    """
    Contains any logic related to scraping of data from discudemy.com
    """

    DOMAIN = "https://discudemy.com"

    def __init__(self, enabled, max_pages=None):
        super().__init__()
        self.scraper_name = "discudemy"
        if not enabled:
            self.set_state_disabled()
        self.max_pages = max_pages

    @BaseScraper.time_run
    async def run(self) -> List:
        """
        Called to gather the udemy links

        :return: List of udemy course links
        """
        links = await self.get_links()
        logger.info(
            f"Page: {self.current_page} of {self.last_page} scraped from discudemy.com"
        )
        self.max_pages_reached()
        return links

    async def get_links(self) -> List:
        """
        Scrape udemy links from discudemy.com

        :return: List of udemy course urls
        """
        discudemy_links = []
        self.current_page += 1
        coupons_data = await get(f"{self.DOMAIN}/all/{self.current_page}")
        soup = BeautifulSoup(coupons_data.decode("utf-8"), "html.parser")
        for course_card in soup.find_all("a", class_="card-header"):
            url_end = course_card["href"].split("/")[-1]
            discudemy_links.append(f"{self.DOMAIN}/go/{url_end}")

        links = await self.gather_udemy_course_links(discudemy_links)

        for counter, course in enumerate(links):
            logger.debug(f"Received Link {counter + 1} : {course}")

        self.last_page = self._get_last_page(soup)

        return links

    @classmethod
    async def get_udemy_course_link(cls, url: str) -> str:
        """
        Gets the udemy course link

        :param str url: The url to scrape data from
        :return: Coupon link of the udemy course
        """

        data = await get(url)
        soup = BeautifulSoup(data.decode("utf-8"), "html.parser")
        for link in soup.find_all("a", href=True):
            udemy_link = cls.validate_coupon_url(link["href"])
            if udemy_link is not None:
                return udemy_link

    async def gather_udemy_course_links(self, courses: List[str]):
        """
        Async fetching of the udemy course links from discudemy.com

        :param list courses: A list of discudemy.com course links we want to fetch the udemy links for
        :return: list of udemy links
        """
        return [
            link
            for link in await asyncio.gather(*map(self.get_udemy_course_link, courses))
            if link is not None
        ]

    @staticmethod
    def _get_last_page(soup: BeautifulSoup) -> int:
        """
        Extract the last page number to scrape

        :param soup:
        :return: The last page number to scrape
        """

        return max(
            [
                int(i.text)
                for i in soup.find("ul", class_="pagination3").find_all("li")
                if i.text.isdigit()
            ]
        )



import asyncio
import json
import logging
from typing import List
from urllib.parse import urlencode

from bs4 import BeautifulSoup

from udemy_enroller.http import get
from udemy_enroller.scrapers.base_scraper import BaseScraper

logger = logging.getLogger("udemy_enroller")


class CoursevaniaScraper(BaseScraper):
    """
    Contains any logic related to scraping of data from coursevania.com
    """

    DOMAIN = "https://coursevania.com"

    def __init__(self, enabled, max_pages=None):
        super().__init__()
        self.scraper_name = "coursevania"
        if not enabled:
            self.set_state_disabled()
        self.last_page = None
        self.max_pages = max_pages
        self._nonce = None

    @BaseScraper.time_run
    async def run(self) -> List:
        """
        Called to gather the udemy links

        :return: List of udemy course links
        """
        links = await self.get_links()
        logger.info(
            f"Page: {self.current_page} of {self.last_page} scraped from coursevania.com"
        )
        self.max_pages_reached()
        return links

    async def get_links(self):
        """
        Scrape udemy links from coursevania.com

        :return: List of udemy course urls
        """
        self.current_page += 1
        await self.load_nonce()
        course_links = await self.get_course_links()

        links = await self.gather_udemy_course_links(course_links)

        for counter, course in enumerate(links):
            logger.debug(f"Received Link {counter + 1} : {course}")

        return links

    async def load_nonce(self) -> None:
        """
        Load the nonce value needed to load the correct page data

        :return: None
        """
        if self._nonce is None:
            response = await get(f"{self.DOMAIN}/courses")
            if response is not None:
                soup = BeautifulSoup(response, "html.parser")
                for script_element in soup.find_all("script"):
                    if "var stm_lms_nonces" in str(script_element):
                        data = json.loads(
                            script_element.string.split(" = ")[1].strip()[:-1]
                        )
                        self._nonce = data.get("load_content")

    async def get_course_links(self) -> List:
        """
        Gets the url of pages which contain the udemy link we want to get

        :return: list of pages on coursevania.com that contain Udemy coupons
        """
        query_params = {
            "offset": self.current_page - 1,
            "template": "courses/grid",
            "args": '{"image_d":"img-480-380","per_row":"4","posts_per_page":"12","class":"archive_grid"}',
            "action": "stm_lms_load_content",
            "sort": "date_high",
            "nonce": self._nonce,
        }
        headers = {
            "Host": "coursevania.com",
            "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:85.0) Gecko/20100101 Firefox/85.0",
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate, br",
            "X-Requested-With": "XMLHttpRequest",
            "Connection": "keep-alive",
            "Referer": "https://coursevania.com/courses/",
            "TE": "Trailers",
        }
        query_string = urlencode(query_params)
        response = await get(
            f"{self.DOMAIN}/wp-admin/admin-ajax.php?{query_string}", headers=headers
        )
        if response is not None:
            json_data = json.loads(response)
            coupons_data = json_data.get("content")
            soup = BeautifulSoup(coupons_data, "html.parser")
            links = soup.find_all("a", class_="heading_font")
            course_links = list({link["href"] for link in links})

            self.last_page = json_data.get("pages")

            return course_links

    @staticmethod
    async def get_udemy_course_link(url: str) -> str:
        """
        Gets the udemy course link

        :param str url: The url to scrape data from
        :return: Coupon link of the udemy course
        """
        text = await get(url)
        if text is not None:
            soup = BeautifulSoup(text.decode("utf-8"), "html.parser")
            udemy_link = (
                soup.find("div", class_="stm-lms-buy-buttons").find("a").get("href")
            )
            return udemy_link

    async def gather_udemy_course_links(self, courses: List[str]):
        """
        Async fetching of the udemy course links from coursevania.com

        :param list courses: A list of coursevania.com course links we want to fetch the udemy links for
        :return: list of udemy links
        """
        return [
            link
            for link in await asyncio.gather(*map(self.get_udemy_course_link, courses))
            if link is not None
        ]



import asyncio
from functools import reduce
from typing import List

from udemy_enroller.scrapers.coursevania import CoursevaniaScraper
from udemy_enroller.scrapers.discudemy import DiscUdemyScraper
from udemy_enroller.scrapers.freebiesglobal import FreebiesglobalScraper
from udemy_enroller.scrapers.tutorialbar import TutorialBarScraper


class ScraperManager:
    def __init__(
        self,
        freebiesglobal_enabled,
        tutorialbar_enabled,
        discudemy_enabled,
        coursevania_enabled,
        max_pages,
    ):
        self.freebiesglobal_scraper = FreebiesglobalScraper(
            freebiesglobal_enabled, max_pages=max_pages
        )
        self.tutorialbar_scraper = TutorialBarScraper(
            tutorialbar_enabled, max_pages=max_pages
        )
        self.discudemy_scraper = DiscUdemyScraper(
            discudemy_enabled, max_pages=max_pages
        )
        self.coursevania_scraper = CoursevaniaScraper(
            coursevania_enabled, max_pages=max_pages
        )
        self._scrapers = (
            self.freebiesglobal_scraper,
            self.tutorialbar_scraper,
            self.discudemy_scraper,
            self.coursevania_scraper,
        )

    async def run(self) -> List:
        """
        Runs any enabled scrapers and returns a list of links

        :return: list
        """
        urls = []
        enabled_scrapers = self._enabled_scrapers()
        if enabled_scrapers:
            urls = reduce(
                list.__add__,
                await asyncio.gather(*map(lambda sc: sc.run(), enabled_scrapers)),
            )
        return urls

    def _enabled_scrapers(self) -> List:
        """
        Returns a list of scrapers that should run

        :return:
        """
        return list(filter(lambda sc: sc.should_run(), self._scrapers))



import aiohttp

from udemy_enroller.logging import get_logger

logger = get_logger()


async def get(url, headers=None):
    """
    Send REST get request to the url passed in

    :param url: The Url to get call get request on
    :param headers: The headers to pass with the get request
    :return: data if any exists
    """
    if headers is None:
        headers = {}
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=headers) as response:
                text = await response.read()
                return text
    except Exception as e:
        logger.error(f"Error in get request: {e}")



import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="udemy-enroller",
    version="4.1.2",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="aapatre",
    author_email="udemyenroller@gmail.com",
    maintainer="fakeid cullzie",
    url="https://github.com/aapatre/Automatic-Udemy-Course-Enroller-GET-PAID-UDEMY-COURSES-for-FREE",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="udemy, education, enroll",
    packages=find_packages(
        exclude=["*tests*"],
    ),
    python_requires=">=3.8, <4",
    install_requires=[
        "aiohttp[speedups]==3.8.1",
        "beautifulsoup4==4.11.1",
        "ruamel.yaml==0.16.13",
        "requests==2.27.1",
        "cloudscraper==1.2.60",
        "webdriver-manager==3.7.0",
        "selenium==3.141.0",
        "price-parser==0.3.4",
    ],
    setup_requires=["pytest-runner"],
    extras_require={
        "dev": ["black", "isort"],
        "test": ["pytest", "pytest-cov"],
    },
    entry_points={
        "console_scripts": [
            "udemy_enroller=udemy_enroller.cli:main",
        ],
    },
)



class RobotException(Exception):
    """
    You have been identified as a robot on Udemy site
    """

    pass


class LoginException(Exception):
    """
    You have failed to login to the Udemy site
    """

    pass



import logging
import logging.config
import os

from udemy_enroller.utils import get_app_dir


class CustomFileHandler(logging.FileHandler):
    """
    Allows us to log to the app directory
    """

    def __init__(self, file_name="app.log", mode="a"):
        log_file_path = os.path.join(get_app_dir(), file_name)
        super(CustomFileHandler, self).__init__(log_file_path, mode)


def load_logging_config() -> None:
    """
    Load logging configuration

    :return: None
    """

    my_logger = logging.getLogger("udemy_enroller")
    my_logger.setLevel(logging.INFO)

    # File handler
    file_handler = CustomFileHandler()
    log_format = "%(asctime)s - %(name)s - %(levelname)s - %(module)s : %(message)s"
    formatter = logging.Formatter(fmt=log_format)
    file_handler.setFormatter(formatter)
    my_logger.addHandler(file_handler)

    # Basic format for streamhandler
    stream_handler = logging.StreamHandler()
    simple_format = logging.Formatter(fmt="%(message)s")
    stream_handler.setFormatter(simple_format)
    my_logger.addHandler(stream_handler)


def get_logger() -> logging.Logger:
    """
    Convenience method to load the app logger

    :return: An instance of the app logger
    """
    return logging.getLogger("udemy_enroller")



import os


def get_app_dir() -> str:
    """
    Gets the app directory where all data related to the script is stored

    :return:
    """
    app_dir = os.path.join(os.path.expanduser("~"), ".udemy_enroller")

    if not os.path.isdir(app_dir):
        # If the app data dir does not exist create it
        os.mkdir(app_dir)
    return app_dir



import asyncio
import random
import time
from typing import Union

from selenium.common.exceptions import (
    NoSuchElementException,
    TimeoutException,
    WebDriverException,
)

from udemy_enroller import (
    ScraperManager,
    Settings,
    UdemyActions,
    UdemyActionsUI,
    UdemyStatus,
    exceptions,
)
from udemy_enroller.logging import get_logger

logger = get_logger()


def _redeem_courses(settings: Settings, scrapers: ScraperManager) -> None:
    """
    Method to scrape courses from the supported sites and enroll in them on udemy

    :param Settings settings: Core settings used for Udemy
    :param ScraperManager scrapers:
    :return:
    """
    udemy_actions = UdemyActions(settings)
    udemy_actions.login()
    loop = asyncio.get_event_loop()

    while True:
        udemy_course_links = loop.run_until_complete(scrapers.run())
        logger.info(f"Total courses this time: {len(udemy_course_links)}")
        if udemy_course_links:
            for course_link in udemy_course_links:
                try:
                    status = udemy_actions.enroll(course_link)

                    if status == UdemyStatus.ENROLLED.value:
                        # Try to avoid udemy throttling by sleeping for 1-5 seconds
                        sleep_time = random.choice(range(1, 5))
                        logger.debug(
                            f"Sleeping for {sleep_time} seconds between enrolments"
                        )
                        time.sleep(sleep_time)
                except KeyboardInterrupt:
                    udemy_actions.stats.table()
                    logger.error("Exiting the script")
                    return
                except Exception as e:
                    logger.error(f"Unexpected exception: {e}")
                finally:
                    if settings.is_ci_build:
                        logger.info("We have attempted to subscribe to 1 udemy course")
                        logger.info("Ending test")
                        return
        else:
            udemy_actions.stats.table()
            logger.info("All scrapers complete")
            return


def redeem_courses(
    settings: Settings,
    freebiesglobal_enabled: bool,
    tutorialbar_enabled: bool,
    discudemy_enabled: bool,
    coursevania_enabled: bool,
    max_pages: Union[int, None],
) -> None:
    """
    Wrapper of _redeem_courses which catches unhandled exceptions

    :param Settings settings: Core settings used for Udemy
    :param bool freebiesglobal_enabled: Boolean signifying if freebiesglobal scraper should run
    :param bool tutorialbar_enabled: Boolean signifying if tutorialbar scraper should run
    :param bool discudemy_enabled: Boolean signifying if discudemy scraper should run
    :param bool coursevania_enabled: Boolean signifying if coursevania scraper should run
    :param int max_pages: Max pages to scrape from sites (if pagination exists)
    :return:
    """
    try:
        scrapers = ScraperManager(
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
            max_pages,
        )
        _redeem_courses(settings, scrapers)
    except Exception as e:
        logger.error(f"Exception in redeem courses: {e}")


def _redeem_courses_ui(
    driver,
    settings: Settings,
    scrapers: ScraperManager,
) -> None:
    """
    Method to scrape courses from the supported sites and enroll in them on udemy.

    :param WebDriver driver: WebDriver to use to complete enrolment
    :param Settings settings: Core settings used for Udemy
    :param ScraperManager scrapers:
    :return:
    """
    udemy_actions = UdemyActionsUI(driver, settings)
    udemy_actions.login()
    loop = asyncio.get_event_loop()

    while True:
        udemy_course_links = loop.run_until_complete(scrapers.run())

        if udemy_course_links:
            for course_link in set(
                udemy_course_links
            ):  # Cast to set to remove duplicate links
                try:
                    status = udemy_actions.enroll(course_link)
                    if status == UdemyStatus.ENROLLED.value:
                        # Try to avoid udemy throttling by sleeping for 1-5 seconds
                        sleep_time = random.choice(range(1, 5))
                        logger.debug(
                            f"Sleeping for {sleep_time} seconds between enrolments"
                        )
                        time.sleep(sleep_time)
                except NoSuchElementException as e:
                    logger.error(f"No such element: {e}")
                except TimeoutException:
                    logger.error(f"Timeout on link: {course_link}")
                except WebDriverException:
                    logger.error(f"Webdriver exception on link: {course_link}")
                except KeyboardInterrupt:
                    udemy_actions.stats.table()
                    logger.warning("Exiting the script")
                    return
                except exceptions.RobotException as e:
                    logger.error(e)
                    return
                except Exception as e:
                    logger.error(f"Unexpected exception: {e}")
                finally:
                    if settings.is_ci_build:
                        logger.info("We have attempted to subscribe to 1 udemy course")
                        logger.info("Ending test")
                        return
        else:
            udemy_actions.stats.table()
            logger.info("All scrapers complete")
            return


def redeem_courses_ui(
    driver,
    settings: Settings,
    freebiesglobal_enabled: bool,
    tutorialbar_enabled: bool,
    discudemy_enabled: bool,
    coursevania_enabled: bool,
    max_pages: Union[int, None],
) -> None:
    """
    Wrapper of _redeem_courses so we always close browser on completion

    :param WebDriver driver: WebDriver to use to complete enrolment
    :param Settings settings: Core settings used for Udemy
    :param bool freebiesglobal_enabled: Boolean signifying if freebiesglobal scraper should run
    :param bool tutorialbar_enabled: Boolean signifying if tutorialbar scraper should run
    :param bool discudemy_enabled: Boolean signifying if discudemy scraper should run
    :param bool coursevania_enabled: Boolean signifying if coursevania scraper should run
    :param int max_pages: Max pages to scrape from sites (if pagination exists)
    :return:
    """

    try:
        scrapers = ScraperManager(
            freebiesglobal_enabled,
            tutorialbar_enabled,
            discudemy_enabled,
            coursevania_enabled,
            max_pages,
        )
        _redeem_courses_ui(driver, settings, scrapers)
    except Exception as e:
        logger.error(f"Exception in redeem courses: {e}")
    finally:
        logger.info("Closing browser")
        driver.quit()



from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.core.utils import ChromeType
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager, IEDriverManager
from webdriver_manager.opera import OperaDriverManager

from udemy_enroller.logging import get_logger

logger = get_logger()

VALID_FIREFOX_STRINGS = {"ff", "firefox"}
VALID_CHROME_STRINGS = {"chrome", "google-chrome"}
VALID_CHROMIUM_STRINGS = {"chromium"}
VALID_INTERNET_EXPLORER_STRINGS = {"internet_explorer", "ie"}
VALID_OPERA_STRINGS = {"opera"}
VALID_EDGE_STRINGS = {"edge"}

ALL_VALID_BROWSER_STRINGS = VALID_CHROME_STRINGS.union(VALID_CHROMIUM_STRINGS)


class DriverManager:
    def __init__(self, browser: str, is_ci_build: bool = False):
        self.driver = None
        self.options = None
        self.browser = browser
        self.is_ci_build = is_ci_build
        self._init_driver()

    def _init_driver(self):
        """
        Initialize the correct web driver based on the users requested browser

        :return: None
        """

        if self.browser.lower() in VALID_CHROME_STRINGS:
            if self.is_ci_build:
                self.options = self._build_ci_options_chrome()

            self.driver = webdriver.Chrome(
                ChromeDriverManager().install(), options=self.options
            )
        elif self.browser.lower() in VALID_CHROMIUM_STRINGS:
            self.driver = webdriver.Chrome(
                ChromeDriverManager(chrome_type=ChromeType.CHROMIUM).install()
            )
        elif self.browser.lower() in VALID_EDGE_STRINGS:
            self.driver = webdriver.Edge(EdgeChromiumDriverManager().install())
        elif self.browser.lower() in VALID_FIREFOX_STRINGS:
            self.driver = webdriver.Firefox(
                executable_path=GeckoDriverManager().install()
            )
        elif self.browser.lower() in VALID_OPERA_STRINGS:
            self.driver = webdriver.Opera(
                executable_path=OperaDriverManager().install()
            )
        elif self.browser.lower() in VALID_INTERNET_EXPLORER_STRINGS:
            self.driver = webdriver.Ie(IEDriverManager().install())
        else:
            raise ValueError("No matching browser found")

        # Get around captcha
        self.driver.execute_cdp_cmd(
            "Page.addScriptToEvaluateOnNewDocument",
            {
                "source": "const newProto = navigator.__proto__;"
                "delete newProto.webdriver;"
                "navigator.__proto__ = newProto;"
            },
        )
        # Maximize the browser
        self.driver.maximize_window()

    @staticmethod
    def _build_ci_options_chrome():
        """
        Build chrome options required to run in CI

        :return:
        """
        # Having the user-agent with Headless param was always leading to robot check
        user_agent = (
            "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 "
            "Safari/537.36"
        )
        options = ChromeOptions()
        # We need to run headless when using github CI
        options.add_argument("--headless")
        options.add_argument("user-agent={0}".format(user_agent))
        options.add_argument("accept-language=en-GB,en-US;q=0.9,en;q=0.8")
        options.add_argument("--window-size=1325x744")
        logger.info("This is a CI run")
        return options



import datetime
import logging
import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional

logger = logging.getLogger("udemy_enroller")


class ScraperStates(Enum):
    DISABLED = "DISABLED"
    RUNNING = "RUNNING"
    COMPLETE = "COMPLETE"


class BaseScraper(ABC):
    def __init__(self):
        self._state = None
        self.scraper_name = None
        self.max_pages = None
        self.last_page = None
        self.current_page = 0

    @abstractmethod
    async def run(self):
        return

    @abstractmethod
    async def get_links(self):
        return

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        if any([ss for ss in ScraperStates if ss.value == value]):
            self._state = value

    def set_state_disabled(self):
        self.state = ScraperStates.DISABLED.value
        logger.info(f"{self.scraper_name} scraper disabled")

    def set_state_running(self):
        self.state = ScraperStates.RUNNING.value
        logger.info(f"{self.scraper_name} scraper is running")

    def set_state_complete(self):
        self.state = ScraperStates.COMPLETE.value
        logger.info(f"{self.scraper_name} scraper complete")

    def is_disabled(self):
        return self.state == ScraperStates.DISABLED.value

    def is_complete(self):
        return self.state == ScraperStates.COMPLETE.value

    def should_run(self):
        should_run = not self.is_disabled() and not self.is_complete()
        if should_run:
            self.set_state_running()
        return should_run

    @staticmethod
    def time_run(func):
        async def wrapper(self):
            start_time = datetime.datetime.utcnow()
            try:
                response = await func(self)
            except Exception as e:
                logger.error(f"Error while running {self.scraper_name} scraper: {e}")
                self.is_complete()
                return []
            end_time = datetime.datetime.utcnow()
            logger.info(
                f"Got {len(response)} links from {self.DOMAIN} in {(end_time - start_time).total_seconds():.2f} seconds"
            )
            return response

        return wrapper

    def max_pages_reached(self) -> bool:
        """
        Returns boolean of whether or not we should continue checking tutorialbar.com

        :return:
        """

        should_run = True

        if self.max_pages is not None:
            should_run = self.max_pages > self.current_page

            if not should_run:
                logger.info(
                    f"Stopping loop. We have reached max number of pages to scrape: {self.max_pages}"
                )
                self.set_state_complete()

        if self.last_page == self.current_page:
            logger.info(
                f"Stopping loop. We have reached the last page to scrape: {self.last_page}"
            )
            self.set_state_complete()

        return should_run

    @staticmethod
    def validate_coupon_url(url) -> Optional[str]:
        """
        Validate the udemy coupon url passed in
        If it matches the pattern it is returned else it returns None

        :param url: The url to check the udemy coupon pattern for
        :return: The validated url or None
        """
        url_pattern = r"^https:\/\/www.udemy.com.*couponCode=.*$"
        matching = re.match(url_pattern, url)
        if matching is not None:
            matching = matching.group()
        return matching
